/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

// nearest neighbor data association

#include "ratbot-slam.h"

// inputs:
//   features         a single feature to match against the map
//   map              the map containing the landmarks we're matching against
//   in_range         indices in the map vector of landmarks to match against
//   n_in_range       # of landmarks to match against
//   correspondences  output map index
//   md_2             output squared mahalanobis distance
//
// a positive output index indicates a valid correspondence was found,
// a negative index means no landmark in the map matched the feature.
// the squared mahalanobis distance to the nearest landmark is stored
// in md_2 (even if it is too large to count as a correspondence)

void nn_data_association
  (const extent_feature *feat, const extent_feature *map,
   const int *in_range, int n_in_range,
   int *correspondence, fixed_t *md_2)
{
  fixed_t m;
  int i;

  *correspondence = -1;
  *md_2 = fp_max_value;

  if(n_in_range <= 0) {
    *correspondence = -1;
    return;
  }

  for(i = 0; i < n_in_range; ++i) {
    if(!segments_overlap(feat, &map[in_range[i]]))
      continue;

    //#if 0
    // FIXME HACK
    if(fp_abs(feat->r - map[in_range[i]].r) > 13107 || angular_distance(feat->theta, map[in_range[i]].theta) > 13107)
      continue;
    //#endif

    m = squared_mahalanobis(feat, &map[in_range[i]]);
    if(m < *md_2 && m < DA_NN_THRESH_2) {
      *correspondence = in_range[i];
      *md_2 = m;
    }
  }

#if 0
  // FIXME REMOVE
  if(*correspondence >= 0) {
    printf("correspondence: %g %g -- %g %g (md2=%g)\n",
	   fp2float(feat->r), fp2float(feat->theta),
	   fp2float(map[*correspondence].r), fp2float(map[*correspondence].theta),
	   fp2float(*md_2));
  }
#endif
}
