/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

// this is a driver program for ratbot-slam that runs slam onboard the
// ratbot, but reads pre-recorded data in from the serial port rather
// than collecting it from the onboard sensors.  mainly useful for
// timing.  send data here with sendlog.

#include "ratbot-slam.h"
#include <stdlib.h>
#include "uart.h"
#include "rprintf.h"
#include "timer128.h"
#include "global.h"
#include "a2d.h"
#include "util.h"
#include "spi.h"
#include "led.h"
#include "sensor.h"
#include "rprintf.h"
#include "buttonInterface.h"

void extmem_init (void) __attribute__ ((naked)) \
   __attribute__ ((section (".init3")));
void extmem_init (void)
{
  MCUCR = 0x80; // enable memory interface
}

int main(void)
{
  timerInit();
  ledInit();
  a2dInit();

  ledBlink(8,1);

  spi_init();
  ir_init();
  buttonInit();
  ledBlink(4,1);

  uartInit();
  uartSetBaudRate(9600);
  rprintfInit((void *)uart_putc);

  ledBlink(2,1);

  ir_set(0x1f); // turn on irs
  fp_srand(readIR(IR_FRONT)); // seed rng using current ir reading
  ir_set(0x00); // turn off irs


#if 0
  // fixed_t testing
  fixed_t a = int2fp(5), b = float2fp(0.5);
  printfp("a", a);
  printfp("b", b);
  fixed_t c = fp_mul(a,b); printfp("mul", c);
  c = fp_div(a,b); printfp("div", c);
  c = a+b; printfp("+", c);
  c = a-b; printfp("-", c);
  c = fp_rand(); printfp("rand", c);
  c = fp_rand(); printfp("rand", c);
  c = fp_tan(a); printfp("tan", c);
  c = fp_cos(a); printfp("cos", c);
  c = fp_sin(a); printfp("sin", c);
  c = fp_atan(a); printfp("atan", c);
  c = fp_atan2(a,b); printfp("atan2", c);
  c = fp_sqrt(a); printfp("sqrt", c);
  c = fp_log(a); printfp("log", c);
  c = fp_exp(b); printfp("exp", c);
#endif


  pose_t start;
  start.x = start.y = start.t = 0;

  filter_init(&start);

  int16_t frame[7];
  int i;
  int done;
  uint32_t startframe, endframe, totframe = 0;

  timer1SetPrescaler(TIMER_CLK_DIV256);


  // testing frame
  frame[0] = 100;
  frame[1] = 100;
  frame[2] = frame[3] = frame[5] = frame[6] = 40;
  frame[4] = 100;

  rprintf("START\r\n");

  int count = 0;

  while(1) {
    done = 1;
#if 0
    for(i = 0; i < 7; ++i) {
      frame[i] = (int16_t)uart_getc() | ((int16_t)uart_getc() << 8);
      if(frame[i] != 0xffff) // 7 0xffff's in a row means we're done
	done = 0;
    }

    if(done)
      break;
#endif

    //    TCNT1 = 0;
    //    startframe = TCNT1;
    filter_update(frame[0], frame[1], frame+2);
    //    endframe = TCNT1;
    //    totframe += endframe - startframe;


//FIXME REMOVE
++count;
if(count > 860)
break;


// FIXME RESTORE
//    uart_putc(0xff); // ready for more!
// rprintf("%d\r\n", count);

  }

  //  rprintfNum(10, 6, 0, ' ', totframe);

  ledBlink(1,1);
  ledBlink(8,1);
  ledBlink(2,1);
  ledBlink(4,1);
  //  while(!getButtonState()); // wait for a button press

  print_filter_state();

  return 0;
}
