/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

#ifndef _RATBOT_SLAM_H
#define _RATBOT_SLAM_H

#include <fixedpoint.h>

//
// parameters and precomputed values
// all of the fixed-point values are for 16.16 representation (the
// default with my fixed point library)
//
// all distances are in METERS and all angles are in RADIANS
//

#define MAX_TRAJECTORY_LENGTH 20000

/*
////// WHAT I USED ON MY LAPTOP
// whole-valued parameters
#define MAX_FEATURES_PER_SCAN 30     // int
#define LE_MIN_POINTS         8      // int
//#define LE_MIN_POINTS         30      // int
//#define N                     13     // int -- number of particles
#define N                     40     // int -- number of particles
//#define M                     40     // int -- multiscan size (sim)
#define M                     55     // int -- multiscan size (ratbot)
#define TRAJ_STORAGE_SIZE     M      // int -- number of historical poses to store
#define MAX_LANDMARKS         50     // int -- maximum map size
*/
// whole-valued parameters
#define MAX_FEATURES_PER_SCAN 30     // int
//#define LE_MIN_POINTS         8      // int
#define LE_MIN_POINTS         8      // int
#define N                     100     // int -- number of particles
#define M                     180     // int -- multiscan size (ratbot)
#define TRAJ_STORAGE_SIZE     M      // int -- number of historical poses to store
#define MAX_LANDMARKS         50     // int -- maximum map size
//#define MAX_LANDMARKS         125


#ifdef _USE_DOUBLE

// double-precision floating point values for all the real-valued
// parameters

#define LE_CLUSTER_THRESH_2   0.0625
#define LE_MIN_LENGTH_2       0.0625
//#define LE_TMAX_2             0.01
#define LE_TMAX_2             0.04
#define LE_MERGE_THRESH_2     9.0

#define VAR_RANGE             0.01
#define VAR_BEARING           0.0012

//#define OVERLAP_EXTRA_LEN     0.5
#define OVERLAP_EXTRA_LEN     0.25
#define IN_RANGE_FOR_DA_2     2.0
#define DA_NN_THRESH_2        9.0
#define NN_THRESH_LIKELIHOOD  0.01110899654

#define one_over_N            (1.0/N)

#define NEFF_MIN              0.5

//#define VAR_ENCODER_PERC      0.0625
#define VAR_ENCODER_PERC      0.02
#define MOTION_c              0.0002086213872
#define MOTION_c_over_2       0.0001043106936
//#define MOTION_c_over_2b      0.0008765604504
#define MOTION_c_over_2b      0.001642688088 //(MOTION_c/2.0/0.0635)

#define CONVERT_IR_RB         0.01
#define CONVERT_IR_RF         0.01
#define CONVERT_IR_F          0.01
#define CONVERT_IR_LF         0.01
#define CONVERT_IR_LB         0.01

#define IR_OFF_X_BACK         0.05675
#define IR_OFF_X_CORNER       0.03675
#define IR_OFF_X_FRONT        0.0585
#define IR_OFF_Y              0.048

#define IR_MIN                0.05
#define IR_MAX                0.8

#else // _USE_DOUBLE

// fixed-point values for all the real-valued parameters

#define LE_CLUSTER_THRESH_2   4096   // fixed_t -- (real=0.25^2, err=0)
//#define LE_MIN_LENGTH_2       4096   // fixed_t -- (real=0.25^2, err=0)
//#define LE_MIN_LENGTH_2 2048
#define LE_MIN_LENGTH_2       655
#define LE_TMAX_2             655    // fixed_t -- (real=0.1^2, err=5.49294e-06)
//#define LE_TMAX_2              163
//#define LE_TMAX_2             2621   // fixed_t -- (real=0.2^2, err=6.71297e-06)
//#define LE_TMAX_2 6000
#define LE_MERGE_THRESH_2     589824 // fixed_t -- (real=3^2, err=0)

#define VAR_RANGE             655    // fixed_t -- std(range)^2 (real=0.1^2, err=5.49294e-06)
#define VAR_BEARING           79     // fixed_t -- std(bearing)^2 (real=0.035^2, err=1.30254e-05)

//#define OVERLAP_EXTRA_LEN     16384  // fixed_t -- allowable non-overlap between two matched features
#define OVERLAP_EXTRA_LEN 6553
//#define OVERLAP_EXTRA_LEN 0
#define IN_RANGE_FOR_DA_2     fp_2   // fixed_t -- cutoff range below which we perform NN
//#define DA_NN_THRESH_2        589824 // fixed_t -- (real=3^2, err=0)
#define DA_NN_THRESH_2        262144 // fixed_t -- (real=2^2, err=0)
// FIXME FIXME:
#define NN_THRESH_LIKELIHOOD  728    // fixed_t -- da_nn_thresh likelihood (real=0.01110899654, err=5.97909e-07)

#define one_over_N            5041   // fixed_t -- 1/N (real=0.07692307692, err=3.52412e-06)

#define NEFF_MIN              32768  // fixed_t -- min. % effective particles (real=0.5, err=0)

// motion model parameters: see my thesis for a complete description
// of the model; briefly:
//   W = wheel diameter (m)
//   C = encoder pulses per revolutions
//   c = pi*W/C
//   b = axle width
// since we want to get rid of as much runtime computation as
// possible, it's convenient to precompute a few simple expressions
// involving these variables.
//#define VAR_ENCODER_PERC 20000
#define VAR_ENCODER_PERC 8192
//#define VAR_ENCODER_PERC      4096   // fixed_t -- % sigma_encoder^2 multiplier (real=0.0625, err=0)
//#define VAR_ENCODER_PERC      2200
//#define VAR_ENCODER_PERC      3000
#define MOTION_c              13     // fixed_t -- pi*W/C (real=0.0002086213872, err=1.02571e-05)
#define MOTION_c_over_2       6      // fixed_t -- c/2 (real=0.0001043106936, err=1.2758e-05)
#define MOTION_c_over_2b      57     // fixed_t -- c/(2*b) (real=0.0008765604504, err=6.80948e-06)
//#define MOTION_c_over_2b      107    // fixed_t -- c/(2*b) (real=0.001642688088, err=9.99763e-06)

// ir sensor parameters -- these may be different per-sensor after
// calibration
#define CONVERT_IR_RB         655    // fixed_t -- 1/100 (ranges are in cm) (real=0.01, err=5.49294e-06)
#define CONVERT_IR_RF         655    // fixed_t -- 1/100
#define CONVERT_IR_F          655    // fixed_t -- 1/100
#define CONVERT_IR_LF         655    // fixed_t -- 1/100
#define CONVERT_IR_LB         655    // fixed_t -- 1/100

#define IR_OFF_X_BACK         3719   // fixed_t -- o_x^back (real=0.05675, err=2.563e-06)
#define IR_OFF_X_CORNER       2408   // fixed_t -- o_x^corner (real=0.03675, err=6.83591e-06)
#define IR_OFF_X_FRONT        3833   // fixed_t -- o_x^front (real=0.0585, err=1.30609e-05)
#define IR_OFF_Y              3145   // fixed_t -- o_y (real=0.048, err=1.11088e-05)

#define IR_MIN                3276   // fixed_t -- rmin (real=0.05, err=1.22078e-05)
#define IR_MAX                52428  // fixed_t -- rmax (real=0.8, err=1.2219e-05)

#endif // _USE_DOUBLE


//
// basic types
//

typedef struct
{
  fixed_t xx, xy, xt,
              yy, yt,
                  tt;
} cov3_t;

typedef struct
{
  fixed_t x, y, t;
} pose_t;

typedef struct
{
  fixed_t r, theta;   // line parameters
  fixed_t rr, rt, tt; // line covariance

  // extent information (not part of the system "state", i.e., not
  // filtered)
  fixed_t umin, xmin, ymin;
  fixed_t umax, xmax, ymax;
} extent_feature;


//
// basic math and geometry
//

fixed_t angular_distance(fixed_t a1, fixed_t a2);

fixed_t squared_distance_point_point(fixed_t x1, fixed_t y1, fixed_t x2, fixed_t y2);
fixed_t squared_distance_point_line(fixed_t x, fixed_t y, fixed_t r, fixed_t theta);
fixed_t squared_distance_point_segment(fixed_t x, fixed_t y, const extent_feature *seg);
fixed_t squared_length(const extent_feature *seg);
fixed_t project_point_onto_line(fixed_t x, fixed_t y, fixed_t r, fixed_t theta);
void u_to_point(fixed_t u, fixed_t r, fixed_t theta, fixed_t *x, fixed_t *y);

void transform_points
  (const fixed_t *x, const fixed_t *y, int n,
   const pose_t *delta, fixed_t *xout, fixed_t *yout);
void transform_points_and_covs
  (const fixed_t *x, const fixed_t *y,
   const fixed_t *xx, const fixed_t *xy, const fixed_t *yy,
   int n, const pose_t *delta,
   fixed_t *xout, fixed_t *yout, fixed_t *xxout, fixed_t *xyout, fixed_t *yyout);

fixed_t squared_mahalanobis(const extent_feature *f1, const extent_feature *f2);
fixed_t likelihood_from_md_2(const extent_feature *f1, const extent_feature *f2, fixed_t md_2);
void transform_segment(const extent_feature *in, const pose_t *delta, extent_feature *out);
int segments_overlap(const extent_feature *f1, const extent_feature *f2);
// stores merged feature distribution in f1
void merge_segments(extent_feature *f1, const extent_feature *f2);

// stores merged pose distribution in p1/c1
void merge_poses(pose_t *p1, cov3_t *c1, const pose_t *p2, const cov3_t *c2);

// adds c2 to c1
void add_cov3(cov3_t *c1, const cov3_t *c2);

// inverts c
void inv_cov3(cov3_t *c);

#define unit_gaussian_shift 6
extern fixed_t unit_gaussian_table[];
fixed_t gaussian(fixed_t mean, fixed_t stddev);
void gaussian_pose(const pose_t *mean, const cov3_t *cov, pose_t *sample);


//
// general slam
//

void extract_line_segments
  (const fixed_t *x, const fixed_t *y, const fixed_t *u, int n,
   extent_feature *out, int *num_features);

void nn_data_association
  (const extent_feature *feat, const extent_feature *map,
   const int *in_range, int n_in_range,
   int *correspondence, fixed_t *md_2);

// updates mu/cov with an ekf prediction
void motion_model(int16_t left, int16_t right, pose_t *mu, cov3_t *cov);

//
// particle filtering
//

// do the necessary filter initialization and set the starting pose
void filter_init(const pose_t *start);

// perform a single step slam update with the given odometry
// information and sensor measurements.  if for some reason there's no
// sensor data for this frame you can set the values in ir[] to
// fp_max_value, which is beyond the maximum range, and they will end
// up being unused.  similarly, if there is no encoder data, set
// left=right=0 and, while a new "pose" will be added to the
// trajectory, the particles will not be projected forward at all.
void filter_update(int16_t left, int16_t right, int16_t ir[5]);

int best_particle(void);

// send the filter state over radio so it can be collected and
// analyzed off-board the robot
void transmit_filter_state(void);

#ifdef _TESTING
#include <stdio.h>
void write_current_particles(FILE *pfile);
void write_best_trajectory(FILE *tfile);
void write_best_map(FILE *mfile);
#endif

// print out state information for the filter
void print_filter_state(void);

#ifndef _TESTING
void printfp(char *pre, fixed_t n);
#endif

#endif // _RATBOT_SLAM_H
